
import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;

public class Comeback64 extends Applet implements WindowListener, Runnable, ImageObserver, MouseListener, MouseMotionListener, KeyListener
{
	private final int palette[] = { 0xFF000000, 0xFFFFFFFF, 0xFFAD524A, 0xFF73F7F7, 0xFFBD6BBD, 0xFF73E773, 0xFF4242C6, 0xFFFFFF7B, 0xFFC88E2F, 0xFF846339, 0xFFFF9C9C, 0xFF7B7B7B, 0xFF949494, 0xFFA5FFA5, 0xFF9C9CF7, 0xFFCECECE };

	private Frame frame = null;
	private int width, height;

	private Thread thread;
	private boolean initialized = false;

	private int pixelBuffer[] = new int[ 320 * 200 ];
	private Image bufferImage;

	private boolean bufferDragged = false;
	private int dragOriginX, dragOriginY, dragCurrentX, dragCurrentY;

	private int scale = 1;
	private int bufferX = 0;
	private int bufferY = 0;
	private boolean drawBorders = true;

	public static void main( String args[] )
	{
		Comeback64 cb64 = new Comeback64();

		cb64.frame = new Frame( "Comeback 64" );
		cb64.frame.addWindowListener( cb64 );
		cb64.frame.add( cb64 );
		cb64.frame.pack();
		cb64.frame.show();
	}

	public Dimension getPreferredSize()
	{
		return new Dimension( 320, 200 );
	}

	public void windowActivated( WindowEvent event ){ requestFocus(); }
	public void windowClosed( WindowEvent event ){ System.exit( 0 ); }
	public void windowClosing( WindowEvent event ){ frame.dispose(); }
	public void windowDeactivated( WindowEvent event ){}
	public void windowDeiconified( WindowEvent event ){}
	public void windowIconified( WindowEvent event ){}
	public void windowOpened( WindowEvent event ){ start(); }

	public void init()
	{
	}

	public void start()
	{
		thread = new Thread( this );
		thread.start();
	}

	public void stop()
	{
		thread.stop();
	}

	public boolean imageUpdate( Image image, int infoflags, int x, int y, int width, int height )
	{
		if( infoflags == ImageObserver.ALLBITS ) return false;
		return true;
	}

	public void paint( Graphics gfx )
	{
		gfx.setColor( Color.black );
		gfx.fillRect( 0, 0, getSize().width, getSize().height );
	}

	public void repaint( Graphics gfx )
	{
	}

	public void run()
	{
		try
		{
			if( !initialized )
			{
				requestFocus();
				addMouseListener( this );
				addMouseMotionListener( this );
				addKeyListener( this );

				bufferImage = createImage( new MemoryImageSource( 320, 200, pixelBuffer, 0, 320 ) );

				if( !initialize() ) return;

				initialized = true;
			}

			while( true )
			{
				byte virtualScreen[] = calculateFrame( 9 );

				Graphics gfx = getGraphics();
				if( gfx != null )
				{
					for( int i=320*200-1; i>=0; i-- )
						pixelBuffer[ i ] = palette[ virtualScreen[ i ] ];

					bufferImage.flush();
					if( scale == 1 ) gfx.drawImage( bufferImage, bufferX, bufferY, this );
					else gfx.drawImage( bufferImage, bufferX, bufferY, 320 * scale, 200 * scale, this );

					if( bufferDragged )
					{
						bufferDragged = false;

						int current = dragCurrentX;
						bufferX += current - dragOriginX;
						dragOriginX = current;

						current = dragCurrentY;
						bufferY += current - dragOriginY;
						dragOriginY = current;

						drawBorders = true;
					}

					if( drawBorders )
					{
						drawBorders = false;

						Dimension dim = getSize();
						int width = dim.width;
						int height = dim.height;

						int xEnd = bufferX + scale * 320;
						int yEnd = bufferY + scale * 200;

						gfx.setColor( Color.black );
						if( bufferX > 0 ) gfx.fillRect( 0, 0, bufferX, height );
						if( bufferY > 0 ) gfx.fillRect( 0, 0, width, bufferY );
						if( xEnd < width ) gfx.fillRect( xEnd, 0, width - xEnd, height );
						if( yEnd < height ) gfx.fillRect( 0, yEnd, width, height - yEnd );
					}

					while( !next ) Thread.sleep( 200 );
					next = go;
				}

			//	Thread.sleep( 40 );
			}
		}
		catch( InterruptedException e )
		{
			System.out.println( "ERROR. Main thread interrupted." );
			if( frame != null ) frame.dispose();
		}
	}

	public void mousePressed( MouseEvent event )
	{
		requestFocus();

		dragOriginX = event.getX();
		dragOriginY = event.getY();
	}

	public void mouseDragged( MouseEvent event )
	{
		dragCurrentX = event.getX();
		dragCurrentY = event.getY();

		bufferDragged = true;
	}

	public void mouseReleased( MouseEvent event ){}
	public void mouseClicked( MouseEvent event ){}
	public void mouseEntered( MouseEvent event ){}
	public void mouseExited( MouseEvent event ){}
	public void mouseMoved( MouseEvent event ){}

	public void keyPressed( KeyEvent event )
	{
		switch( event.getKeyChar() )
		{
			case '+':
				scale += 1;
				return;

			case '-':
				if( scale == 1 ) break;
				scale -= 1;
				drawBorders = true;
				return;

			case 'n':
				next = true;
				return;

			case 'g':
				next = go = true;
				return;
		}
	}

	public void keyReleased( KeyEvent event )
	{
		switch( event.getKeyCode() )
		{
		}
	}

	public void keyTyped( KeyEvent event ){}

















	private boolean next = true;
	private boolean go = true;



	private boolean negative_flag = false;
	private boolean overflow_flag = false;
	private boolean break_flag = true;
	private boolean decimal_flag = false;
	private boolean int_flag = false;
	private boolean zero_flag = false;
	private boolean carry_flag = false;



	private int joytoemu = 1;

	private byte basicMem[] = new byte[ 8192 ];
	private byte kernalMem[] = new byte[ 8192 ];
	private byte normalMem[] = new byte[ 65536 ];
	private byte charMem[] = new byte[ 4096 ];
	private byte colorMem[] = new byte[ 1000 ];
	private byte virtualScreen[] = new byte[ 65536 ];

	private byte vicRegs[] = new byte[ 0x2F ];
	private byte sidRegs[] = new byte[ 0x1D ];
	private byte cia1Regs[] = new byte[ 0x10 ];
	private byte cia2Regs[] = new byte[ 0x10 ];

	boolean enable1a, enable1b, enable2a, enable2b, vic_den;
	byte vic_ec, vic_b0c;

	int vic_y, vic_vm, vic_cb, vic_bank;
	int latch1a=0, latch2a=0, latch1b=0, latch2b=0;
	byte cpuport, vic_yscrl, vic_xscrl, vic_b1c, vic_b2c, vic_b3c;
	boolean vic_ecm, vic_bmm, vic_rsel, vic_mcm, vic_csel;
	boolean irq = false;
	byte icr1set=0, icr2set=0;
	int t1a, t2a, t1b, t2b;

	int curr_loc;
	boolean nmi = false;
	int vic_rasirq = 0xFFFF;

	byte stack_ptr = (byte)0xFF; 
	byte a_reg=0,x_reg=0,y_reg=0,gen1;

	int address, jmp_from;

	int skip_cycles = 0;

	private final int readFile( String name, byte dest[], int off, int len )
	throws Exception
	{
		InputStream stream;

		if( frame != null ) stream = new FileInputStream( new File( name ) );
		else stream = new URL( getCodeBase(), name ).openStream();

		stream = new BufferedInputStream( stream );
		int bytes = stream.read( dest, off, len );
		stream.close();

		return bytes;
	}

	private final void loadROM( String name, byte dest[] )
	throws Exception
	{
		if( readFile( "roms/" + name + ".rom", dest, 0, dest.length ) != dest.length )
			throw new Exception();
	}

	private final boolean initialize()
	{
		try
		{
			loadROM( "basic", basicMem );
			loadROM( "kernal", kernalMem );
			loadROM( "char", charMem );
		}
		catch( Exception e )
		{
			System.out.println( "ERROR. Couldn't load rom files." );
			return false;
		}

		clearMem();

		curr_loc = ( 0xFF00 & ( kernalMem[ 0x1FFD ] << 8 ) ) + ( 0xFF & (int)kernalMem[ 0x1FFC ] );

		return true;
	}

	private final void clearMem()
	{
		for( int i=0; i<65535; i++ )
		{
			if( ( i & 0x80 ) == 0 ) normalMem[ i ] = (byte)0x99;
			else normalMem[ i ] = 0x66;
		}

		for( int i=0; i<1000; i++ ) colorMem[ i ] = 0;
		for( int i=0; i<0x2F; i++ ) vicRegs[ i ] = 0;
		for( int i=0; i<0x1D; i++ ) sidRegs[ i ] = 0;
		for( int i=0; i<0x10; i++ ) cia1Regs[ i ] = cia2Regs[ i ] = 0;

		normalMem[ 0 ] = 0x2F;
		normalMem[ 1 ] = 0x37;

		enable1a = enable1b = enable2a = enable2b = vic_den = false;
		vic_ec = vic_b0c = 0;
	}

	private final byte getFlags()
	{
		byte flags = 32;

		if( negative_flag ) flags += 128;
		if( overflow_flag ) flags += 64;
		if( break_flag ) flags += 16;
		if( decimal_flag ) flags += 8;
		if( int_flag ) flags += 4;
		if( zero_flag ) flags += 2;
		if( carry_flag ) flags += 1;

		return flags;
	}

	private final void setFlags( byte flags )
	{
		negative_flag = flags < 0;
		overflow_flag = ( flags & 64 ) != 0;
		break_flag = ( flags & 16 ) != 0;
		decimal_flag = ( flags & 8 ) != 0;
		int_flag = ( flags & 4 ) != 0;
		zero_flag = ( flags & 2 ) != 0;
		carry_flag = ( flags & 1 ) != 0;
	}

	private final byte[] calculateFrame( int skipFrames )
	{
		while( skipFrames-- >= 0 )
		{
			int vic_x = vic_y = 0;

			boolean frameDone = false;

			while( !frameDone )
			{
				vic_x += 8;

				if( vic_x >= 506 )
				{
					vic_x = 0;

					if( ++vic_y == 312 )
					{
						if( skipFrames < 0 )
						{
						//	CIA();
						//	Sprite();
						}

						vic_y = 0;
						frameDone = true;
					}

					if( vic_y == vic_rasirq )
					{
						vicRegs[ 0x19 ] |= 1;
						if( ( vicRegs[ 0x1A ] & 0x1 ) != 0 )
						{
							irq = true;
							vicRegs[ 0x19 ] |= 0x80;
						}
					}
				}

				if( skip_cycles > 0 )
				{
					skip_cycles--;
				}
				else
				{
					boolean skip = false;

					if( irq )
					{
						if( !int_flag )
						{
							poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), (byte)( curr_loc >> 8 ) );
							poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), (byte)curr_loc );
							poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), (byte)( getFlags() & 239 ) );
							curr_loc = deek( 0xFFFE );
							int_flag = true;
						}
						else skip = true;
					}

					if( nmi && !skip )
					{
						nmi = false;

						poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), (byte)( curr_loc >> 8 ) );
						poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), (byte)curr_loc );
						poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), getFlags() );
						curr_loc = deek( 0xFFFA );
					}

					switch( 0xFF & peek( curr_loc++ ) )
					{
						case 0x00:	// BRK();
							poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), (byte)( ( curr_loc + 1 ) >> 8 ) );
							poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), (byte)( curr_loc + 1 ) );
							poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), getFlags() );
							int_flag = true;
							curr_loc = deek( 0xFFFE );
							skip_cycles = 5;
							break;

						case 0x01:	// ORA_IndX();
							a_reg |= peek( deek( 0xFF & ( peek( curr_loc++ ) + x_reg ) ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 4;
							break;

						case 0x05:	// ORA_Zero();
							a_reg |= peek( 0xFF & peek( curr_loc++ ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 1;
							break;

						case 0x06:	// ASL_Zero();
							address = 0xFF & peek( curr_loc++ );
							gen1 = peek( address );
							carry_flag = gen1 < 0;
							gen1 <<= 1;
							poke( address, gen1 );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 3;
							break;

						case 0x08:	// PHP();
							poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), getFlags() );
							skip_cycles = 1;
							break;

						case 0x09:	// ORA_Imm();
							a_reg |= peek( curr_loc++ );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							break;

						case 0x0A:	// ASL_Acc();
							carry_flag = a_reg < 0;
							a_reg <<= 1;
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							break;

						case 0x0D:	// ORA_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;

							a_reg |= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							skip_cycles = 2;
							break;

						case 0x0E:	// ASL_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							gen1 = peek( address );
							carry_flag = gen1 < 0;
							gen1 <<= 1;
							poke( address, gen1 );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 4;
							break;

						case 0x10:	// BPL();
							if( !negative_flag )
							{
								skip_cycles = 1;
								curr_loc += gen1 = peek( curr_loc );
							}
							curr_loc++;
							break;

						case 0x11:	// ORA_Ind_Y();
							a_reg |= peek( deek( 0xFF & peek( curr_loc++ ) ) + ( 0xFF & y_reg ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 3;
							break;

						case 0x15:	// ORA_Zero_X();
							a_reg |= peek( 0xFF & ( peek( curr_loc++ ) + x_reg ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 2;
							break;

						case 0x16:	// ASL_Zero_X();
							address = 0xFF & ( peek( curr_loc++ ) + x_reg );
							gen1 = peek( address );
							carry_flag = gen1 < 0;
							gen1 <<= 1;
							poke( address, gen1 );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 4;
							break;

						case 0x18:	// CLC();
							carry_flag = false;
							break;

						case 0x19:	// ORA_Abs_Y();
							address = deek( curr_loc ) + ( 0xFF & y_reg );
							curr_loc += 2;
							a_reg |= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							if( ( ( address - ( 0xFF & (int)y_reg ) ) / 256 ) == ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 4;
							break;

						case 0x1D:	// ORA_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							a_reg |= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							if( ( ( address - ( 0xFF & (int)x_reg ) ) / 256 ) == ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 4;
							break;

						case 0x1E:	// ASL_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							gen1 = peek( address );
							carry_flag = gen1 < 0;
							gen1 <<= 1;
							poke( address, gen1 );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 5;
							break;

						case 0x20:	// JSR();
						        jmp_from = ++curr_loc;
							poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), (byte)( curr_loc >> 8 ) );
							poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), (byte)curr_loc );
							curr_loc = deek( curr_loc - 1 );
							break;

						case 0x21:	// AND_IndX();
							a_reg &= peek( deek( 0xFF & ( peek( curr_loc++ ) + x_reg ) ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 4;
							break;

						case 0x24:	// BIT_Zero();
							gen1 = peek( 0xFF & peek( curr_loc++ ) );
							setflags( (byte)( a_reg & gen1 ) );
							negative_flag = ( gen1 & 0x80 ) != 0;
							overflow_flag = ( gen1 & 0x40 ) != 0;
							skip_cycles = 1;
							break;

						case 0x25:	// AND_Zero();
							a_reg &= peek( 0xFF & peek( curr_loc++ ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 1;
							break;

						case 0x26:	// ROL_Zero();
							address = 0xFF & peek( curr_loc++ );

							gen1 = (byte)( peek( address ) << 1 );
							if( carry_flag ) gen1 |= 1;
							else gen1 &= 0xFE;
							carry_flag = ( peek( address ) & 0x80 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 3;
							break;

						case 0x28:	// PLP();
							setFlags( (byte)( peek( 0x100 + ( 0xFF & (int)(++stack_ptr) ) ) | 0x30 ) );
							skip_cycles = 2;
							break;

						case 0x29:	// AND_Imm();
							a_reg &= peek( curr_loc++ );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							break;

						case 0x2A:	// ROL_Acc();
							gen1 = (byte)( a_reg << 1 );
							if( carry_flag ) gen1 |= 1;
							carry_flag = ( a_reg & 0x80 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							a_reg = gen1;
							break;

						case 0x2C:	// BIT_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							gen1 = peek( address );
							setflags( (byte)( a_reg & gen1 ) );
							negative_flag = ( gen1 & 0x80 ) != 0;
							overflow_flag = ( gen1 & 0x40 ) != 0;
							skip_cycles = 2;
							break;

						case 0x2D:	// AND_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;

							a_reg &= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							skip_cycles = 2;
							break;

						case 0x2E:	// ROL_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;

							gen1 = (byte)( peek( address ) << 1 );
							if( carry_flag ) gen1 |= 1;
							else gen1 &= 0xFE;
							carry_flag = ( peek( address ) & 0x80 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 4;
							break;

						case 0x30:	// BMI();
							if( negative_flag )
							{
								curr_loc += gen1 = peek( curr_loc );
								skip_cycles = 1;
							}
							curr_loc++;
							break;

						case 0x31:	// AND_Ind_Y();
							address = deek( 0xFF & peek( curr_loc++ ) ) + ( 0xFF & y_reg );
							a_reg &= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 3;
							break;

						case 0x35:	// AND_Zero_X();
							address = 0xFF & ( peek( curr_loc++ ) + x_reg );
							a_reg &= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 2;
							break;

						case 0x36:	// ROL_Zero_X();
							address = 0xFF & ( peek( curr_loc++ ) + x_reg );

							gen1 = (byte)( peek( address ) << 1 );
							if( carry_flag ) gen1 |= 1;
							else gen1 &= 0xFE;
							carry_flag = ( peek( address ) & 0x80 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 3;
							break;

						case 0x38:	// SEC();
							carry_flag = true;
							break;

						case 0x39:	// AND_Abs_Y();
							address = deek( curr_loc ) + ( 0xFF & y_reg );
							curr_loc += 2;
							a_reg &= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							if( ( ( address - ( 0xFF & (int)y_reg ) ) / 256 ) == ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 4;
							break;

						case 0x3D:	// AND_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							a_reg &= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							if( ( ( address - (int)( 0xFF & x_reg ) ) / 256 ) == ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 4;
							break;

						case 0x3E:	// ROL_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;

							gen1 = (byte)( peek( address ) << 1 );
							if( carry_flag ) gen1 |= 1;
							else gen1 &= 0xFE;
							carry_flag = ( peek( address ) & 0x80 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 4;
							break;

						case 0x40:	// RTI();
							setFlags( (byte)( peek( 0x100 + ( 0xFF & (int)(++stack_ptr) ) ) | 0x30 ) );
							curr_loc = 0xFF & (int)peek( 0x100 + ( 0xFF & (int)(++stack_ptr) ) );
							curr_loc += 0xFF00 & ( peek( 0x100 + ( 0xFF & (int)(++stack_ptr) ) ) << 8 );
							skip_cycles = 4;
							break;

						case 0x41:	// EOR_IndX();
							address	= 0xFF & ( peek( curr_loc++ ) + x_reg );
							address = deek( address );
							a_reg ^= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 4;
							break;

						case 0x45:	// EOR_Zero();
							address = 0xFF & peek( curr_loc++ );
							a_reg ^= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 1;
							break;

						case 0x46:	// LSR_Zero();
							address = 0xFF & peek( curr_loc++ );

							gen1 = peek( address );
							carry_flag = ( gen1 & 1 ) != 0;
							gen1 >>= 1;
							gen1 &= 0x7F;
							negative_flag = false;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 3;
							break;

						case 0x48:	// PHA();
							poke( 0x100 + ( 0xFF & (int)stack_ptr-- ), a_reg );
							skip_cycles = 1;
							break;

						case 0x49:	// EOR_Imm();
							a_reg ^= peek( curr_loc++ );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							break;

						case 0x4A:	// LSR_Acc();
							carry_flag = ( a_reg & 1 ) != 0;
							a_reg >>= 1;
							a_reg &= 0x7F;
							negative_flag = false;
							zero_flag = a_reg == 0;
							break;

						case 0x4C:	// JMP_Abs();
							jmp_from = curr_loc - 1;
							curr_loc = deek( curr_loc );
							skip_cycles = 1;
							break;

						case 0x4D:	// EOR_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							a_reg ^= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 2;
							break;

						case 0x4E:	 // LSR_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;

							gen1 = peek( address );
							carry_flag = ( gen1 & 1 ) != 0;
							gen1 >>= 1;
							gen1 &= 0x7F;
							negative_flag = false;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 4;
							break;

						case 0x50:	// BVC();
							if( !overflow_flag )
							{
								curr_loc += gen1 = peek( curr_loc );
								skip_cycles = 1;
							}
							curr_loc++;
							break;

						case 0x51:	// EOR_Ind_Y();
							a_reg ^= peek( deek( 0xFF & peek( curr_loc++ ) ) + ( 0xFF & y_reg ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 3;
							break;

						case 0x55:	// EOR_Zero_X();
							a_reg ^= peek( 0xFF & ( peek( curr_loc++ ) + x_reg ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 2;
							break;

						case 0x56:	// LSR_Zero_X();
							address = 0xFF & ( peek( curr_loc++ ) + x_reg );

							gen1 = peek( address );
							carry_flag = ( gen1 & 1 ) != 0;
							gen1 >>= 1;
							gen1 &= 0x7F;
							negative_flag = false;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 4;
							break;

						case 0x58:	// CLI();
							int_flag = false;
							break;

						case 0x59:	// EOR_Abs_Y();
							address = deek( curr_loc ) + ( 0xFF & y_reg );
							curr_loc += 2;
							a_reg ^= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							if( ( ( address - ( 0xFF & (int) y_reg ) ) / 256 ) == ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 4;
							break;

						case 0x5D:	// EOR_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							a_reg ^= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							if( ( ( address - ( 0xFF & (int)x_reg ) ) / 256 ) == ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 4;
							break;

						case 0x5E:	// LSR_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;

							gen1 = peek( address );
							carry_flag = ( gen1 & 1 ) != 0;
							gen1 >>= 1;
							gen1 &= 0x7F;
							negative_flag = false;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 4;
							break;

						case 0x60:	// RTS();
							curr_loc = 0xFF & (int)peek( 0x100 + ( 0xFF & (int)(++stack_ptr) ) );
							curr_loc += 0xFF00 & ( peek( 0x100 + ( 0xFF & (int)(++stack_ptr) ) ) << 8 );
							curr_loc++;
							skip_cycles = 4;
							break;

						case 0x61:	// ADC_IndX();
							address	= 0xFF & ( peek( curr_loc++ ) + x_reg );
							address = deek( address );
					        	ADC();
							skip_cycles = 4;
							break;

						case 0x65:	// ADC_Zero();
							address = 0xFF & peek( curr_loc++ );
							ADC();
							skip_cycles = 1;
							break;

						case 0x66:	// ROR_Zero();
							address = 0xFF & peek( curr_loc++ );

							gen1 = (byte)( peek( address ) >> 1 );
							if( carry_flag ) gen1 |= 0x80;
							else gen1 &= 0x7F;
							carry_flag = ( peek( address ) & 1 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 3;
							break;

						case 0x68:	// PLA();
							a_reg = peek( 0x100 + ( 0xFF & (int)(++stack_ptr) ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 2;
							break;

						case 0x69:	// ADC_Imm();
							address = curr_loc++;
							ADC();
							break;

						case 0x6A:	// ROR_Acc();
							gen1 = (byte)( a_reg >> 1 );
							if( carry_flag ) gen1 |= 0x80;
							carry_flag = ( a_reg & 1 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							a_reg = gen1;
							break;

						case 0x6C:	// JMP_Ind();
							jmp_from = curr_loc - 1;
							address = deek( curr_loc );
							int k = address;
							curr_loc = 0xFF & (int)peek( k );
							k++;
							if( ( k & 0xFF ) == 0 ) k -= 0x100;
							curr_loc += (int)( peek( k ) << 8 );
							skip_cycles = 3;
							break;

						case 0x6D:	// ADC_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							ADC();
							skip_cycles = 2;
							break;

						case 0x6E:	// ROR_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;

							gen1 = (byte)( peek( address ) >> 1 );
							if( carry_flag ) gen1 |= 0x80;
							else gen1 &= 0x7F;
							carry_flag = ( peek( address ) & 1 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 4;
							break;

						case 0x70:	// BVS();
							if( overflow_flag )
							{
								curr_loc += gen1 = peek( curr_loc );
								skip_cycles = 1;
							}
							curr_loc++;
							break;

						case 0x71:	// ADC_Ind_Y();
							address = deek( 0xFF & peek( curr_loc++ ) ) + ( 0xFF & y_reg );
							ADC();
							skip_cycles = 3;
							break;

						case 0x75:	// ADC_Zero_X();
							address = 0xFF & ( peek( curr_loc++ ) + x_reg );
							ADC();
							skip_cycles = 2;
							break;

						case 0x76:	// ROR_Zero_X();
							address = 0xFF & ( peek( curr_loc++ ) + x_reg );

							gen1 = (byte)( peek( address ) >> 1 );
							if( carry_flag ) gen1 |= 0x80;
							else gen1 &= 0x7F;
							carry_flag = ( peek( address ) & 1 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 3;
							break;

						case 0x78:	// SEI();
							int_flag = true;
							break;

						case 0x79:	// ADC_Abs_Y();
							address = deek( curr_loc ) + ( 0xFF & y_reg );
							curr_loc += 2;
							ADC();

							if( ( ( address - ( 0xFF & (int)y_reg ) ) / 256 ) == ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 4;
							break;

						case 0x7D:	// ADC_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							ADC();

							if( ( ( address - ( 0xFF & (int)x_reg ) ) / 256 ) == ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 4;
							break;

						case 0x7E:	// ROR_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;

							gen1 = (byte)( peek( address ) >> 1 );
							if( carry_flag ) gen1 |= 0x80;
							else gen1 &= 0x7F;
							carry_flag = ( peek( address ) & 1 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							skip_cycles = 4;
							break;

						case 0x81:	// STA_IndX();
							poke( deek( 0xFF & ( peek( curr_loc++ ) + x_reg ) ), a_reg );
							skip_cycles = 4;
							break;

						case 0x84:	// STY_Zero();
							poke( 0xFF & peek( curr_loc++ ), y_reg );
							skip_cycles = 1;
							break;

						case 0x85:	// STA_Zero();
							poke( 0xFF & peek( curr_loc++ ), a_reg );
							skip_cycles = 1;
							break;

						case 0x86:	// STX_Zero();
							poke( 0xFF & peek( curr_loc++ ), x_reg );
							skip_cycles = 1;
							break;

						case 0x88:	// DEY();
							y_reg--;
							negative_flag = y_reg < 0;
							zero_flag = y_reg == 0;
							break;

						case 0x8A:	// TXA();
							a_reg = x_reg;
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							break;

						case 0x8C:	// STY_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							poke( address, y_reg );
							skip_cycles = 2;
							break;

						case 0x8D:	// STA_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							poke( address, a_reg );
							skip_cycles = 2;
							break;

						case 0x8E:	// STX_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							poke( address, x_reg );
							skip_cycles = 2;
							break;

						case 0x90:	// BCC();
							if( !carry_flag )
							{
								curr_loc += gen1 = peek( curr_loc );
								skip_cycles = 1;
							}
							curr_loc++;
							break;

						case 0x91:	// STA_Ind_Y();
							poke( deek( 0xFF & peek( curr_loc++ ) ) + ( 0xFF & y_reg ), a_reg );
							skip_cycles = 4;
							break;

						case 0x94:	// STY_Zero_X();
							poke( 0xFF & ( peek( curr_loc++ ) + x_reg ), y_reg );
							skip_cycles = 2;
							break;

						case 0x95:	// STA_Zero_X();
							poke( 0xFF & ( peek( curr_loc++ ) + x_reg ), a_reg );
							skip_cycles = 2;
							break;

						case 0x96:	// STX_Zero_Y();
							poke( 0xFF & ( peek( curr_loc++ ) + y_reg ), x_reg );
							skip_cycles = 2;
							break;

						case 0x98:	// TYA();
							a_reg = y_reg;
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							break;

						case 0x99:	// STA_Abs_Y();
							poke( deek( curr_loc ) + ( 0xFF & y_reg ), a_reg );
							curr_loc += 2;
							skip_cycles = 3;
							break;

						case 0x9A:	// TXS();
							stack_ptr = x_reg;
							break;

						case 0x9D:	// STA_Abs_X();
							poke( deek( curr_loc ) + ( 0xFF & x_reg ), a_reg );
							curr_loc += 2;
							skip_cycles = 3;
							break;

						case 0xA0:	// LDY_Imm();
							y_reg = peek( curr_loc++ );
							negative_flag = y_reg < 0;
							zero_flag = y_reg == 0;
							break;

						case 0xA1:	// LDA_IndX();
							a_reg = peek( deek( 0xFF & ( peek( curr_loc++ ) + x_reg ) ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 4;
							break;

						case 0xA2:	// LDX_Imm();
							x_reg = peek( curr_loc++ );
							negative_flag = x_reg < 0;
							zero_flag = x_reg == 0;
							break;

						case 0xA4:	// LDY_Zero();
							y_reg = peek( 0xFF & peek( curr_loc++ ) );
							negative_flag = y_reg < 0;
							zero_flag = y_reg == 0;
							skip_cycles = 1;
							break;

						case 0xA5:	// LDA_Zero();
							a_reg = peek( 0xFF & peek( curr_loc++ ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 1;
							break;

						case 0xA6:	// LDX_Zero();
							x_reg = peek( 0xFF & peek( curr_loc++ ) );
							negative_flag = x_reg < 0;
							zero_flag = x_reg == 0;
							skip_cycles = 1;
							break;

						case 0xA8:	// TAY();
							y_reg = a_reg;
							negative_flag = y_reg < 0;
							zero_flag = y_reg == 0;
							break;

						case 0xA9:	// LDA_Imm();
							a_reg = peek( curr_loc++ );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							break;

						case 0xAA:	// TAX();
							x_reg = a_reg;
							negative_flag = x_reg < 0;
							zero_flag = x_reg == 0;
							break;

						case 0xAC:	// LDY_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							y_reg = peek( address );
							negative_flag = y_reg < 0;
							zero_flag = y_reg == 0;
							skip_cycles = 2;
							break;

						case 0xAD:	// LDA_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							a_reg = peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 2;
							break;

						case 0xAE:	// LDX_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							x_reg = peek( address );
							negative_flag = x_reg < 0;
							zero_flag = x_reg == 0;
							skip_cycles = 2;
							break;

						case 0xB0:	// BCS();
							if( carry_flag )
							{
								curr_loc += gen1 = peek( curr_loc );
								skip_cycles = 1;
							}
							curr_loc++;
							break;

						case 0xB1:	// LDA_Ind_Y();
							address = deek( 0xFF & peek( curr_loc++ ) ) + ( 0xFF & y_reg );
							a_reg = peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							if( ( ( address - ( 0xFF & (int)y_reg ) ) / 256 ) != ( address / 256 ) )
								skip_cycles = 4;
							else skip_cycles = 3;
							break;

						case 0xB4:	// LDY_Zero_X();
							y_reg = peek( 0xFF & ( peek( curr_loc++ ) + x_reg ) );
							negative_flag = y_reg < 0;
							zero_flag = y_reg == 0;
							skip_cycles = 2;
							break;

						case 0xB5:	// LDA_Zero_X();
							a_reg = peek( 0xFF & ( peek( curr_loc++ ) + x_reg ) );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 2;
							break;

						case 0xB6:	// LDX_Zero_Y();
							x_reg = peek( 0xFF & ( peek( curr_loc++ ) + y_reg ) );
							negative_flag = x_reg < 0;
							zero_flag = x_reg == 0;
							skip_cycles = 2;
							break;

						case 0xB8:	// CLV();
							overflow_flag = false;
							break;

						case 0xB9:	// LDA_Abs_Y();
							address = deek( curr_loc ) + ( 0xFF & y_reg );
							curr_loc += 2;
							a_reg = peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							if( ( ( address - ( 0xFF & (int)y_reg ) ) / 256 ) != ( address / 256 ) )
								skip_cycles = 4;
							else skip_cycles = 3;
							break;

						case 0xBA:	// TSX();
							x_reg = stack_ptr;
							negative_flag = x_reg < 0;
							zero_flag = x_reg == 0;
							break;

						case 0xBC:	// LDY_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							y_reg = peek( address );
							negative_flag = y_reg < 0;
							zero_flag = y_reg == 0;

							if( ( ( address - ( 0xFF & (int)x_reg ) ) / 256 ) != ( address / 256 ) )
								skip_cycles = 4;
							else skip_cycles = 3;
							break;

						case 0xBD:	// LDA_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							a_reg = peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							if( ( ( address - ( 0xFF & (int)x_reg ) ) / 256 ) != ( address / 256 ) )
								skip_cycles = 4;
							else skip_cycles = 3;
							break;

						case 0xBE:	// LDX_Abs_Y();
							address = deek( curr_loc ) + ( 0xFF & y_reg );
							curr_loc += 2;
							x_reg = peek( address );
							negative_flag = x_reg < 0;
							zero_flag = x_reg == 0;

							if( ( ( address - ( 0xFF & (int)y_reg ) ) / 256 ) != ( address / 256 ) )
								skip_cycles = 4;
							else skip_cycles = 3;
							break;

						case 0xC0:	// CPY_Imm();
							gen1 = peek( curr_loc++ );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)y_reg );
							setflags( (byte)( y_reg - gen1 ) );
							break;

						case 0xC1:	// CMP_IndX();
							address	= 0xFF & ( peek( curr_loc++ ) + x_reg );
							address = deek( address );

							gen1 = peek( address );
							setflags( (byte)( a_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)a_reg );

							skip_cycles = 4;
							break;

						case 0xC4:	// CPY_Zero();
							address = 0xFF & peek( curr_loc++ );

							gen1 = peek( address );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)y_reg );
							setflags( (byte)( y_reg - gen1 ) );

							skip_cycles = 2;
							break;

						case 0xC5:	// CMP_Zero();
							gen1 = peek( 0xFF & peek( curr_loc++ ) );
							setflags( (byte)( a_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)a_reg );
							skip_cycles = 1;
							break;

						case 0xC6:	// DEC_Zero();
							address = 0xFF & peek( curr_loc++ );
							poke( address, gen1 = (byte)( peek( address ) - 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 3;
							break;

						case 0xC8:	// INY();
							y_reg++;
							negative_flag = y_reg < 0;
							zero_flag = y_reg == 0;
							break;

						case 0xC9:	// CMP_Imm();
							gen1 = peek( curr_loc++ );
							setflags( (byte)( a_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)a_reg );

							break;

						case 0xCA:	// DEX();
							x_reg--;
							negative_flag = x_reg < 0;
							zero_flag = x_reg == 0;
							break;

						case 0xCC:	// CPY_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;

							gen1 = peek( address );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)y_reg );
							setflags( (byte)( y_reg - gen1 ) );

							skip_cycles = 2;
							break;

						case 0xCD:	// CMP_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;

							gen1 = peek( address );
							setflags( (byte)( a_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)a_reg );

							skip_cycles = 2;
							break;

						case 0xCE:	// DEC_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							poke( address, gen1 = (byte)( peek( address ) - 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 4;
							break;

						case 0xD0:	// BNE();
							if( !zero_flag )
							{
								curr_loc += gen1 = peek( curr_loc );
								skip_cycles = 1;
							}
							curr_loc++;
							break;

						case 0xD1:	// CMP_Ind_Y();
							address = deek( 0xFF & peek( curr_loc++ ) ) + ( 0xFF & y_reg );
							gen1 = peek( address );
							setflags( (byte)( a_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)a_reg );

							if( ( ( address - ( 0xFF & (int)y_reg ) ) / 256 ) != ( address / 256 ) )
								skip_cycles = 4;
							else skip_cycles = 3;
							break;

						case 0xD5:	// CMP_Zero_X();
							address = 0xFF & ( peek( curr_loc++ ) + x_reg );

							gen1 = peek( address );
							setflags( (byte)( a_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)a_reg );

							skip_cycles = 2;
							break;

						case 0xD6:	// DEC_Zero_X();
							address = 0xFF & ( peek( curr_loc++ ) + x_reg );
							poke( address, gen1 = (byte)( peek(address) - 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 4;
							break;

						case 0xD8:	// CLD();
							decimal_flag = false;
							break;

						case 0xD9:	// CMP_Abs_Y();
							address = deek( curr_loc ) + ( 0xFF & y_reg );
							curr_loc += 2;

							gen1 = peek( address );
							setflags( (byte)( a_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)a_reg );

							if( ( ( address - ( 0xFF & (int)y_reg ) ) / 256 ) != ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 2;
							break;

						case 0xDD:	// CMP_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							gen1 = peek( address );
							setflags( (byte)( a_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)a_reg );

							if( ( ( address - ( 0xFF & (int)x_reg ) ) / 256 ) != ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 2;
							break;

						case 0xDE:	// DEC_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							poke( address, gen1 = (byte)( peek(address) - 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 5;
							break;

						case 0xE0:	// CPX_Imm();
							gen1 = peek( curr_loc++ );
							setflags( (byte)( x_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)x_reg );
							break;

						case 0xE1:	// SBC_IndX();
							address	= 0xFF & ( peek( curr_loc++ ) + x_reg );
							address = deek( address );
							SBC();
							skip_cycles = 4;
							break;

						case 0xE4:	// CPX_Zero();
							address = 0xFF & peek( curr_loc++ );

							gen1 = peek( address );
							setflags( (byte)( x_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)x_reg );

							skip_cycles = 1;
							break;

						case 0xE5:	// SBC_Zero();
							address = 0xFF & peek( curr_loc++ );
							SBC();
							skip_cycles = 1;
							break;

						case 0xE6:	// INC_Zero();
							address = 0xFF & peek( curr_loc++ );
							poke( address, gen1 = (byte)( peek( address ) + 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 3;
							break;

						case 0xE8:	// INX();
							x_reg++;
							negative_flag = x_reg < 0;
							zero_flag = x_reg == 0;
							break;

						case 0xE9:	// SBC_Imm();
							address = curr_loc++;
							SBC();
							break;

						case 0xEC:	// CPX_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;

							gen1 = peek( address );
							setflags( (byte)( x_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)x_reg );

							skip_cycles = 2;
							break;

						case 0xED:	// SBC_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;
							SBC();
							skip_cycles = 2;
							break;

						case 0xEE:	// INC_Abs();
							address = 0xFF & peek( curr_loc++ );
							address += ( 0xFF & peek( curr_loc++ ) ) << 8;

							poke( address, gen1 = (byte)( peek(address) + 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;

							skip_cycles = 4;
							break;

						case 0xF0:	// BEQ();
							if( zero_flag )
							{
								curr_loc += gen1 = peek( curr_loc );
								skip_cycles = 1;
							}
							curr_loc++;
							break;

						case 0xF1:	// SBC_Ind_Y();
							address = deek( 0xFF & peek( curr_loc++ ) ) + ( 0xFF & y_reg );
							SBC();
							skip_cycles = 3;
							break;

						case 0xF5:	// SBC_Zero_X();
							address = 0xFF & ( peek( curr_loc++ ) + x_reg );
							SBC();
							skip_cycles = 2;
							break;

						case 0xF6:	// INC_Zero_X();
							address = 0xFF & ( peek( curr_loc++ ) + x_reg );
							poke( address, gen1 = (byte)( peek( address ) + 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 4;
							break;

						case 0xF8:	// SED();
							decimal_flag = true;
							break;

						case 0xF9:	// SBC_Abs_Y();
							address = deek( curr_loc ) + ( 0xFF & y_reg );
							curr_loc += 2;
							SBC();

							if( ( ( address - ( 0xFF & (int)y_reg ) ) / 256 ) == ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 4;
							break;

						case 0xFD:	// SBC_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							SBC();

							if( ( ( address - ( 0xFF & (int)x_reg ) ) / 256 ) == ( address / 256 ) )
								skip_cycles = 3;
							else skip_cycles = 4;
							break;

						case 0xFE:	// INC_Abs_X();
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							poke( address, gen1 = (byte)( peek( address ) + 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							skip_cycles = 5;
							break;

						case 0x82:	// SKB
						case 0x80:
						case 0xc2:
						case 0xe2:
						case 0x04:
						case 0x14:
						case 0x34:
						case 0x44:
						case 0x54:
						case 0x64:
						case 0x74:
						case 0x89:	// op tests call it NOP #imm ??
						case 0xD4:
						case 0xf4:
							curr_loc++;
							break;

						case 0x0B:	// ASO_Imm
							gen1 = peek( curr_loc++ );
							if( ( gen1 & 0x80 ) != 0 ) carry_flag = true;
							gen1 <<= 1;
							a_reg |= gen1;
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							skip_cycles = 3;
							break;

						case 0x1F:
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;

							carry_flag = ( peek( address ) & 0x80 ) != 0;
							poke( address, gen1 = (byte)( peek( address ) << 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;

							a_reg |= peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;

							skip_cycles = 7;
							break;

						case 0x67:
							address = 0xFF & peek( curr_loc++ );

							gen1 = (byte)( peek( address ) >> 1 );
							if( carry_flag ) gen1 |= 0x80;
							else gen1 &= 0x7F;
							carry_flag = ( peek( address ) & 1 ) != 0;
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;
							poke( address, gen1 );

							ADC();
							skip_cycles = 4;
							break;

						case 0x87:
							address = 0xFF & peek( curr_loc++ );
							poke( address, (byte)( a_reg & x_reg ) );
							skip_cycles = 4;
							break;

						case 0xBF:
							address = deek( curr_loc ) + ( 0xFF & y_reg );
							curr_loc += 2;
							a_reg = x_reg = peek( address );
							negative_flag = a_reg < 0;
							zero_flag = a_reg == 0;
							break;

						case 0xC3:
							address	= 0xFF & (int)( peek( curr_loc++ ) + x_reg );
							address = deek( address );

							poke( address, gen1 = (byte)( peek( address ) - 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;

							gen1 = peek( address );
							setflags( (byte)( a_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)a_reg );

							skip_cycles = 7;
							break;

						case 0xC7:
							address = 0xFF & peek( curr_loc++ );

							poke( address, gen1 = (byte)( peek( address ) - 1 ) );
							negative_flag = gen1 < 0;
							zero_flag = gen1 == 0;

							gen1 = peek( address );
							setflags( (byte)( a_reg - gen1 ) );
							carry_flag = ( 0xFF & (int)gen1 ) <= ( 0xFF & (int)a_reg );

							skip_cycles = 5;
							break;

						case 0xFF:
							address = deek( curr_loc ) + ( 0xFF & x_reg );
							curr_loc += 2;
							poke( address, (byte)( peek(address) + 1 ) );
							SBC();
							skip_cycles = 7;
							break;

						case 0x0C:	// SKW
						case 0x1C:
						case 0x3C:
						case 0x5C:
						case 0x7C:
						case 0xDC:
						case 0xFC:
							curr_loc += 2;
							break;
					}
				}

				if( enable1a )
				{
					if( t1a == 0 )
					{ 
						t1a = 0xFF & ( cia1Regs[ 4 ] = (byte)latch1a );

						if( ( cia1Regs[ 0xE ] & 0x08 ) != 0 ) cia1Regs[ 0xE ] &= 254;

						if( ( icr1set & 0x01 ) != 0 )
						{
							irq = true;
							cia1Regs[ 0xD ] = (byte)0x80;
						}

						cia1Regs[ 0xD ] |= 1;

						if( ( cia1Regs[ 0xF ] & 0x60 ) == 0x40 ) cia1Regs[ 6 ] = (byte)( --t1b );
					}

					cia1Regs[ 4 ] = (byte)( --t1a );
				}

				if( enable1b )
				{
					if( t1b == 0 )
					{
						t1b = 0xFF & ( cia1Regs[ 6 ] = (byte)latch1b );

						if( ( cia1Regs[ 0xF ] & 0x08 ) != 0 ) cia1Regs[ 0xF ] &= 254;

						if( ( icr1set & 0x02 ) != 0 )
						{
							irq = true;
							cia1Regs[ 0xD ] = (byte)0x80;
						}

						cia1Regs[ 0xD ] |= 2;
					}
					else
					{
						if( ( cia1Regs[ 0xF ] & 0x60 ) == 0 )
						{
							cia1Regs[ 6 ] = (byte)( --t1b );
						}
					}
				}

				if( enable2a )
				{
					if( t2a == 0 )
					{
						t2a = 0xFF & ( cia2Regs[ 4 ] = (byte)latch2a );

						if( ( cia2Regs[ 0xE ] & 0x08 ) != 0 ) cia2Regs[ 0xE ] &= 254;

						if( ( icr2set & 0x01 ) != 0 )
						{
							nmi = true;
							cia2Regs[ 0xD ] = (byte)0x80;
						}

						cia2Regs[ 0xD ] |= 1;

						if( ( cia2Regs[ 0xF ] & 0x60 ) == 0x40 ) cia2Regs[ 0x6 ] = (byte)( --t2b );
					}

					cia2Regs[ 0x4 ] = (byte)( --t2a );
				}

				if( enable2b )
				{
					if( t2b == 0 )
					{
						t2b = 0xFF & ( cia2Regs[ 6 ] = (byte)latch2b );

						if( ( cia2Regs[ 0xF ] & 0x08 ) != 0 ) cia2Regs[ 0xF ] &= 254;

						if( ( icr2set & 0x02 ) != 0 )
						{
							nmi = true;
							cia2Regs[ 0xD ] = (byte)0x80;
						}

						cia2Regs[ 0xD ] |= 2;
					}
					else if( ( cia2Regs[ 0xF ] & 0x60 ) == 0 ) cia2Regs[ 0x6 ] = (byte)( --t2b );
				}
			}
		}

		drawChars();

		return virtualScreen;
	}

	private final int putInst( int loc, int y )
	{
		putCBMStr( hexToString( loc, 4 ), 0, y, 1, 0 );

		putCBMStr( hexToString( peek( loc ), 2 ), 5, y, 2, 0 );
		putCBMStr( hexToString( peek( loc + 1 ), 2 ), 8, y, 2, 0 );
		putCBMStr( hexToString( peek( loc + 2 ), 2 ), 11, y, 2, 0 );

		String str = instNames[ 0xFF & peek( loc ) ];

		int i = str.indexOf( '*' );
		if( i >= 0 )
		{
			str = str.substring( 0, i ) + hexToString( loc + 2 + peek( loc + 1 ), 4 ) + str.substring( i + 2 );
			i = 2;
		}
		else
		{
			i = str.indexOf( 'a' );
			if( i >= 0 )
			{
				str = str.substring( 0, i ) + hexToString( peek( loc + 1 ), 2 ) + str.substring( i + 1 );

				i = str.indexOf( 'b' );
				if( i >= 0 )
				{
					str = str.substring( 0, i ) + hexToString( peek( loc + 2 ), 2 ) + str.substring( i + 1 );
					i = 3;
				}
				else i = 2;
			}
			else i = 1;
		}

		putCBMStr( str, 14, y, 1, 0 );

		return i;
	}

	private final String hexToString( int val, int len )
	{
		String str = Integer.toHexString( val ).toUpperCase();

		if( str.length() > len ) str = str.substring( str.length() - len );
		else while( str.length() < len ) str = "0" + str;

		return str;
	}

	private final void drawChars()
	{
		int pos = 0x0400;

		for( int y=0; y<25; y++ )
			for( int x=0; x<40; x++ )
				putCBMChar( 0xFF & normalMem[ pos++ ], x, y, 14, 6 );
	}

	private final void setflags( byte r )
	{
		negative_flag = r < 0;
		zero_flag = r == 0;
	}

	private final void ADC()	// KOLLA DEN HAR!
	{
		gen1 = peek( address );

		int c;
		if( carry_flag ) c = 1;
		else c = 0;

		if( decimal_flag )
		{
			int k = ( a_reg & 0xF ) + ( gen1 & 0xF ) + c;

			if( ( k & 0xFF ) >= 0xA ) k = ( k & 0xFF00 ) + ( ( k + 0x6 ) & 0xFF );
			if( ( k & 0xFF ) >= 0x20 ) k = ( k & 0xFF0F ) + 0x10;

			k += ( a_reg & 0xF0 ) + ( gen1 & 0xF0 );

			negative_flag = ( k & 0x80 ) != 0;
			zero_flag = ( 0xFF & ( a_reg + gen1 + c ) ) == 0;
			overflow_flag = ( ~( ( 0xFF & (int)a_reg ) ^ ( 0xFF & (int)gen1 ) ) & ( ( 0xFF & (int)a_reg ) ^ ( 0xFF & k ) ) & 0x80 ) != 0;

			if( carry_flag = k >= 0xA0 )
				k += 0x60;

			a_reg = (byte)k;
		}
		else
		{
			int k = ( 0xFF & (int)a_reg ) + ( 0xFF & (int)gen1 ) + c;

			negative_flag = ( k & 0x80 ) != 0;
			zero_flag = ( k & 0xFF ) == 0;
			overflow_flag = ( ~( ( 0xFF & (int)a_reg ) ^ ( 0xFF & (int)gen1 ) ) & ( ( 0xFF & (int)a_reg ) ^ ( 0xFF & k ) ) & 0x80 ) != 0;
			carry_flag = ( k & 0xFF00 ) != 0;

			a_reg = (byte)k;
		}
	}

	private final void SBC()	// KOLLA DEN HAR OCKSA!
	{
		gen1 = peek( address ); 

		int c;
		if( !carry_flag ) c = 1;
		else c = 0;

		if( decimal_flag )
		{
			int k = ( a_reg & 0xF ) + 0x100;
			k -= ( gen1 & 0xF ) + c;
			if( k < 0x100 ) k -= 6;
			if( k < 0xF0 ) k += 0x10;
			k += a_reg & 0xF0;
			k -= gen1 & 0xF0;

			negative_flag = ( k & 0x80 ) != 0;
			zero_flag = a_reg - gen1 - c == 0;
			overflow_flag = ( ~( ( 0xFF & k ) ^ gen1 ) & ( a_reg ^ gen1 ) & 0x80 ) != 0;

			carry_flag = ( k & 0xFF00 ) != 0;
			if( !carry_flag ) k = ( k & 0xFF00 ) + ( 0xFF & ( ( k & 0xFF ) - 0x60 ) );

			a_reg = (byte)k;
		}
		else
		{
			int k = 0x100 + ( 0xFF & (int)a_reg );
			k -= ( 0xFF & (int)gen1 ) + c;

			negative_flag = carry_flag = zero_flag = overflow_flag = false;

			negative_flag = ( k & 0x80 ) != 0;
			zero_flag = ( k & 0xFF ) == 0;
			overflow_flag = ( ~( ( 0xFF & k ) ^ gen1 ) & ( a_reg ^ gen1 ) & 0x80 ) != 0;
			carry_flag = ( k & 0xFF00 ) != 0;

			a_reg = (byte)k;
		}
	}

	private final byte peek( int address )
	{
		address &= 0xFFFF;

		if( address < 0xA000 ) return normalMem[ address ];

		if( address < 0xC000 )
		{
			if( ( normalMem[ 1 ] & 3 ) == 0 ) return normalMem[ address ];
			return basicMem[ address & 0x1FFF ];
		}

		if( address < 0xD000 ) return normalMem[ address ];

		if( address >= 0xE000 )
		{
			if( ( normalMem[ 1 ] & 2 ) == 0 ) return normalMem[ address ];
			return kernalMem[ address & 0x1FFF ];
		}

		byte flags = normalMem[ 1 ];
		if( ( flags & 3 ) == 0 ) return normalMem[ address ];
		if( ( flags & 4 ) == 0 ) return charMem[ address ];

		if( address < 0xD400 )	// VIC
		{
			address &= 0x3F;

			if( address < 0x2F )
			{
				byte val = vicRegs[ address ];

				switch( address )
				{
					case 0x11:
						if( ( vic_y & 0x100 ) != 0 ) val = (byte)( val | 0x80 );
						else val &= 0x7F;
						break;

					case 0x12:
						val = (byte)vic_y;
						break;

					case 0x19:
						val |= 6;
						break;
				}

				return val;
			}

			return -1; //0xFF
		}

		if( address < 0xD800 )	// SID
		{
			address &= 0x1F; 

			if( address < 0x1D ) return sidRegs[ address ];
			return -1; //0xFF
		}

		if( address < 0xDBE8 )	// Color RAM
			return (byte)( colorMem[ address & 0x3FF ] & 0xF );

		if( address < 0xDC00 )	// Nothing
			return -1; //0xFF

		if( address < 0xDD00 )	// CIA 1
		{
			address &= 0xF;
			byte val = cia1Regs[ address ];

			switch( address )
			{
				case 0:
				//	if( joytoemu == 1 )
				//		return ( val & 224 ) + joyemu();
					break;

				case 1:
				//	val = key_pressed[ 255 - cia1Regs[ 0 ] ];
				//	if( joytoemu == 2 && val == 255 )
				//		return ( val & 224 ) + joyemu();
					break;

				case 13:
					cia1Regs[ 13 ] = 0;
					irq = false;
					break;
			}

			return val;
		}

		if( address < 0xDE00 ) // CIA 2
		{
			address &= 0xF;
			byte val = cia2Regs[ address ];

			switch( address )
			{
				case 0:
					val = cia2Regs[ 0 ] = 0;
					break;

				case 13:
					cia2Regs[ 13 ] = 0;
					irq = false;
					break;
			}

			return val;
		}

		if( address < 0xDFA0 )	// Nothing or CIA2 echoes
			return -1; //0xFF

		return 0;
	}

	private final int deek( int address )
	{
		return ( 0xFF & (int)peek( address ) ) + ( 0xFF00 & ( (int)peek( ( address + 1 ) & 0xFFFF ) << 8 ) );
	}

	private final void poke( int address, byte val )
	{
		address &= 0xFFFF;

		if( address < 0xD000 )
		{
			switch( address )
			{
				case 1:
					cpuport = val;
					val = normalMem[ 0 ];
					// no break on purpose 

				case 0:
					normalMem[ 0 ] = val;
					normalMem[ 1 ] &= ~( val | 0x20 );
					normalMem[ 1 ] |= ( cpuport & val ) | ( 0x17 & ~val );
					break;

				default:
					normalMem[ address ] = val;
			}
		}
		else if( address >= 0xE000 ) normalMem[ address ] = val;
		else
		{
			if( ( normalMem[ 1 ] & 7 ) == 0 ) normalMem[ address ] = val;
			else
			{
				if( address < 0xD400 )	// VIC
				{
					address &= 0x3F;

					switch( address )
					{
						case 0x11:
							if( ( val & 0x40 ) != 0 ) vic_ecm = true;
							else vic_ecm = false;
							if( ( val & 0x20 ) != 0 ) vic_bmm = true;
							else vic_bmm = false;
							if( ( val & 0x10 ) != 0 ) vic_den = true;
							else vic_den = false;
							if( ( val & 8 ) != 0 ) vic_rsel = true;
							else vic_rsel = false;
							vic_yscrl = (byte)( val & 7 );
							vic_rasirq = 0xFF & (int)vicRegs[ 0x12 ];
							if( ( val & 0x80 ) != 0 ) vic_rasirq |= 0x100;
							break;

						case 0x12:
							vic_rasirq = 0xFF & (int)val;
							if( ( vicRegs[ 0x11 ] & 0x80 ) != 0 ) vic_rasirq |= 0x100;
							break;

						case 0x16:
							val |= 0xA0;
							if( ( val & 0x10 ) != 0 ) vic_mcm = true;
							else vic_mcm = false;
							if( ( val & 8 ) != 0 ) vic_csel = true;
							else vic_csel = false;
							vic_xscrl = (byte)( val & 7 );
							break;

						case 0x18:
							val |= 1;
							vic_vm = ( (int)val & 0xF0 ) << 6; 
							vic_cb = ( (int)val & 0xE ) << 10;
							break;

						case 0x19:
							val = (byte)( ~( val & 0xF ) & vicRegs[ 0x19 ] );
							if( ( ( val & vicRegs[ 0x1A ] & 0xF ) == 0 ) && ( ( val & 0x80 ) != 0 ) )
							{ 
								irq = false;
								val &= 0x7F;
							}
							val |= 0x70;
							break;

						case 0x1A:
							val |= 0xF0;
							break;

						case 0x20:
							vic_ec = (byte)( val & 0xF );
							val |= 0xF0;
							break;

						case 0x21:
							vic_b0c = (byte)( val & 0xF );
							val |= 0xF0;
							break;

						case 0x22:
							vic_b1c = (byte)( val & 0xF );
							val |= 0xF0;
							break;

						case 0x23:
							vic_b2c = (byte)( val & 0xF );
							val |= 0xF0;
							break;

						case 0x24:
							vic_b3c = (byte)( val & 0xF );
							val |= 0xF0;
							break;
					} 

					if( address < 0x2F ) vicRegs[ address ] = val;
				}
				else if( address < 0xD800 )	// SID
				{
					address &= 0x1F;

					if( address < 0x1F ) sidRegs[ address ] = val;
				}
				else if( address < 0xDBE8 )	// Color RAM
					colorMem[ address & 0x3FF ] = (byte)( val & 0xF );
				else if( address < 0xDC00 );	// Nothing, Johan thinks
				else if( address < 0xDD00 )	// CIA 1
				{
					address &= 0xF;
					cia1Regs[ address ] = val;

					switch( address )
					{ 
						case 4:
							latch1a &= 0xFF00;
							latch1a |= val;
							break;

						case 5:
							latch1a &= 0xFF;
							latch1a |= (int)val << 8;
							break;

						case 6:
							latch1b &= 0xFF00;
							latch1b |= val;
							break;

						case 7:
							latch1b &= 0xFF;
							latch1b |= (int)val << 8;

						case 13:
							if( ( val & 0x01 ) != 0 )
							{
								if( ( val & 0x80 ) != 0 ) icr1set |= 1;
								else icr1set &= 254;
							}
							if( ( val & 0x02 ) != 0 )
							{
								if( ( val & 0x80 ) != 0 ) icr1set |= 2;
								else icr1set &= 253;
							}
							if( ( val & 0x04 ) != 0 )
							{
								if( ( val & 0x80 ) != 0 ) icr1set |= 4;
								else icr1set &= 251;
							}
							if( ( val & 0x08 ) != 0 )
							{
								if( ( val & 0x80 ) != 0 ) icr1set |= 8;
								else icr1set &= 247;
							}
							if( ( val & 0x10 ) != 0 )
							{
								if( ( val & 0x80 ) != 0 ) icr1set |= 16;
								else icr1set &= 239;
							}
							break;

						case 14:
							if( ( val & 0x10 ) != 0 )
							{
								t1a = 0xFF & ( cia1Regs[ 4 ] = (byte)latch1a );
								if( ( val & 0x01 ) != 0 ) enable1a = true;
								else enable1a = false;
							}
							break;

						case 15:
							if( ( val & 0x10 ) != 0 )
							{
								t1b = 0xFF & ( cia1Regs[ 6 ] = (byte)latch1b );
								if( ( val & 0x1 ) != 0 ) enable1b = true;
								else enable1b = false;
							}
							break;
					}
				}
				else if( address < 0xDE00 )	// CIA 2
				{
					address &= 0xF;
					cia2Regs[ address ] = val;

					switch( address )
					{
						case 0:
							vic_bank = ( ~val & 3 ) << 14 ;
							break;

						case 4:
							latch2a &= 0xFF00;
							latch2a |= val;
							break;

						case 5:
							latch2a &= 0xFF;
							latch2a |= (int)val << 8;
							break;

						case 6:
							latch2b &= 0xFF00;
							latch2b |= val;
							break;

						case 7:
							latch2b &= 0xFF;
							latch2b |= (int)val << 8;

						case 13:
							if( ( val & 0x01 ) != 0 )
							{
								if( ( val & 0x80 ) != 0 ) icr2set |= 1;
								else icr2set &= 254;
							}
							if( ( val & 0x02 ) != 0 )
							{
								if( ( val & 0x80 ) != 0 ) icr2set |= 2;
								else icr2set &= 253;
							}
							if( ( val & 0x04 ) != 0 )
							{
								if( ( val & 0x80 ) != 0 ) icr2set |= 4;
								else icr2set &= 251;
							}
							if( ( val & 0x08 ) != 0 )
							{
								if( ( val & 0x80 ) != 0 ) icr2set |= 8;
								else icr2set &= 247;
							}
							if( ( val & 0x10 ) != 0 )
							{
								if( ( val & 0x80 ) != 0 ) icr2set |= 16;
								else icr2set &= 239;
							}
							break;

						case 14:
							if( ( val & 0x10 ) != 0 )
							{
								t2a = 0xFF & ( cia2Regs[ 4 ] = (byte)latch2a );
								if( ( val & 1 ) != 0 ) enable2a = true;
								else enable2a = false;
							}
							break;

						case 15:
							if( ( val & 0x10 ) != 0 )
							{
								t2b = 0xFF & ( cia2Regs[ 6 ] = (byte)latch2b );
								if( ( val & 1 ) != 0 ) enable2b = true;
								else enable2b = false;
							}
							break;
					}
				}
			}
		}
	} 

	private final byte joyemu()
	{
		return 31;
	}

	private final void putCBMChar( int letter, int xp, int yp, int fore, int back )
	{
		byte foreb = (byte)fore;
		byte backb = (byte)back;

		int place = xp * 8 + 320 * yp * 8;

		for( int y=0; y<8; y++ )
		{
			int o = 0xFF & charMem[ letter * 8 + y ];
			int pos = 128;

			for( int x=0; x<8; x++ )
			{
				byte col;

				if( o >= pos )
				{
					o -= pos;
					col = foreb;
				}
				else col = backb;

				virtualScreen[ x + 320 * y + place ] = col;

				pos = pos >> 1;
			}
		}
	}

	private final void putCBMStr( String string, int xp, int yp, int fore, int back )
	{
		int len = string.length();
		if( xp + len > 40 ) len = 40 - xp;

		for( int i=0; i<len; i++ )
			putCBMChar( string.charAt( i ) & 63, xp++, yp, fore, back );
	}


	private final String instNames[] =
	{
/*0x00*/ "BRK", 
/*0x01*/ "ORA ($a,X)", 
/*0x02*/ "???", 
/*0x03*/ "??? ($a,X)", 
/*0x04*/ "??? $a", 
/*0x05*/ "ORA $a", 
/*0x06*/ "ASL $a", 
/*0x07*/ "??? $a", 
/*0x08*/ "PHP", 
/*0x09*/ "ORA #$a", 
/*0x0A*/ "ASL", 
/*0x0B*/ "??? #$a", 
/*0x0C*/ "??? $ba", 
/*0x0D*/ "ORA $ba", 
/*0x0E*/ "ASL $ba", 
/*0x0F*/ "??? $ba", 
/*0x10*/ "BPL $*a", 
/*0x11*/ "ORA ($a),Y", 
/*0x12*/ "???", 
/*0x13*/ "??? ($a),Y", 
/*0x14*/ "??? $a,X", 
/*0x15*/ "ORA $a,X", 
/*0x16*/ "ASL $a,X", 
/*0x17*/ "??? $a,X", 
/*0x18*/ "CLC", 
/*0x19*/ "ORA $ba,Y", 
/*0x1A*/ "???", 
/*0x1B*/ "??? $ba,Y", 
/*0x1C*/ "??? $a", 
/*0x1D*/ "ORA $ba,X", 
/*0x1E*/ "ASL $ba,X", 
/*0x1F*/ "??? $ba,X", 
/*0x20*/ "JSR $ba", 
/*0x21*/ "AND ($a,X)", 
/*0x22*/ "NOP", 
/*0x23*/ "??? ($a,X)", 
/*0x24*/ "BIT $a", 
/*0x25*/ "AND $a", 
/*0x26*/ "ROL $a", 
/*0x27*/ "??? $a", 
/*0x28*/ "PLP", 
/*0x29*/ "AND #$a", 
/*0x2A*/ "ROL", 
/*0x2B*/ "??? #$a", 
/*0x2C*/ "BIT $ba", 
/*0x2D*/ "AND $ba", 
/*0x2E*/ "ROL $ba", 
/*0x2F*/ "??? $ba", 
/*0x30*/ "BMI $*a", 
/*0x31*/ "AND ($a),Y", 
/*0x32*/ "???", 
/*0x33*/ "??? ($a),Y", 
/*0x34*/ "??? $a,X", 
/*0x35*/ "AND $a,X", 
/*0x36*/ "ROL $a,X", 
/*0x37*/ "??? $a,X", 
/*0x38*/ "SEC", 
/*0x39*/ "AND $ba,Y", 
/*0x3A*/ "???", 
/*0x3B*/ "??? $ba,Y", 
/*0x3C*/ "??? $ba,X", 
/*0x3D*/ "AND $ba,X", 
/*0x3E*/ "ROL $ba,X", 
/*0x3F*/ "??? $ba,X", 
/*0x40*/ "RTI", 
/*0x41*/ "EOR ($a,X)", 
/*0x42*/ "???", 
/*0x43*/ "??? ($a,X)", 
/*0x44*/ "??? $a", 
/*0x45*/ "EOR $a", 
/*0x46*/ "LSR $a", 
/*0x47*/ "??? $a", 
/*0x48*/ "PHA", 
/*0x49*/ "EOR #$a", 
/*0x4A*/ "LSR", 
/*0x4B*/ "??? #$a", 
/*0x4C*/ "JMP $ba", 
/*0x4D*/ "EOR $ba", 
/*0x4E*/ "LSR $ba", 
/*0x4F*/ "??? $ba", 
/*0x50*/ "BVC $*a", 
/*0x51*/ "EOR ($a),Y", 
/*0x52*/ "???", 
/*0x53*/ "??? ($a),Y", 
/*0x54*/ "??? $a,X", 
/*0x55*/ "EOR $a,X", 
/*0x56*/ "LSR $a,X", 
/*0x57*/ "??? $a,X", 
/*0x58*/ "CLI", 
/*0x59*/ "EOR $ba,Y", 
/*0x5A*/ "???", 
/*0x5B*/ "??? $ba,Y", 
/*0x5C*/ "??? $ba,X", 
/*0x5D*/ "EOR $ba,X", 
/*0x5E*/ "LSR $ba,X", 
/*0x5F*/ "??? $ba,X", 
/*0x60*/ "RTS", 
/*0x61*/ "ADC ($a,X)", 
/*0x62*/ "???", 
/*0x63*/ "??? ($a,X)", 
/*0x64*/ "??? $a", 
/*0x65*/ "ADC $a", 
/*0x66*/ "ROR $a", 
/*0x67*/ "??? $a", 
/*0x68*/ "PLA", 
/*0x69*/ "ADC #$a", 
/*0x6A*/ "ROR", 
/*0x6B*/ "??? #$a", 
/*0x6C*/ "JMP ($ba)", 
/*0x6D*/ "ADC $ba", 
/*0x6E*/ "ROR $ba", 
/*0x6F*/ "??? $ba", 
/*0x70*/ "BVS $*a", 
/*0x71*/ "ADC ($a),Y", 
/*0x72*/ "???", 
/*0x73*/ "??? ($a),Y", 
/*0x74*/ "??? $a,X", 
/*0x75*/ "ADC $a,X", 
/*0x76*/ "ROR $a,X", 
/*0x77*/ "??? $a,X", 
/*0x78*/ "SEI", 
/*0x79*/ "ADC $ba,Y", 
/*0x7A*/ "???", 
/*0x7B*/ "??? $ba,Y", 
/*0x7C*/ "??? $ba,X", 
/*0x7D*/ "ADC $ba,X", 
/*0x7E*/ "ROR $ba,X", 
/*0x7F*/ "??? $ba,X", 
/*0x80*/ "??? #$a", 
/*0x81*/ "STA ($a,X)", 
/*0x82*/ "??? #$a", 
/*0x83*/ "??? ($a,X)", 
/*0x84*/ "STY $a", 
/*0x85*/ "STA $a", 
/*0x86*/ "STX $a", 
/*0x87*/ "??? $a", 
/*0x88*/ "DEY", 
/*0x89*/ "??? #$a", 
/*0x8A*/ "TXA", 
/*0x8B*/ "??? #$a", 
/*0x8C*/ "STY $ba", 
/*0x8D*/ "STA $ba", 
/*0x8E*/ "STX $ba", 
/*0x8F*/ "??? $ba", 
/*0x90*/ "BCC $*a", 
/*0x91*/ "STA ($a),Y", 
/*0x92*/ "???", 
/*0x93*/ "??? ($a),Y", 
/*0x94*/ "STY $a,X", 
/*0x95*/ "STA $a,X", 
/*0x96*/ "STX $a,Y", 
/*0x97*/ "??? $a,Y", 
/*0x98*/ "TYA", 
/*0x99*/ "STA $ba,Y", 
/*0x9A*/ "TXS", 
/*0x9B*/ "??? $ba,Y", 
/*0x9C*/ "??? $ba,X", 
/*0x9D*/ "STA $ba,X", 
/*0x9E*/ "??? $ba,Y", 
/*0x9F*/ "??? $ba,Y", 
/*0xA0*/ "LDY #$a", 
/*0xA1*/ "LDA ($a,X)", 
/*0xA2*/ "LDX #$a", 
/*0xA3*/ "??? ($a,X)", 
/*0xA4*/ "LDY $a", 
/*0xA5*/ "LDA $a", 
/*0xA6*/ "LDX $a", 
/*0xA7*/ "??? $a", 
/*0xA8*/ "TAY", 
/*0xA9*/ "LDA #$a", 
/*0xAA*/ "TAX", 
/*0xAB*/ "??? #$a", 
/*0xAC*/ "LDY $ba", 
/*0xAD*/ "LDA $ba", 
/*0xAE*/ "LDX $ba", 
/*0xAF*/ "??? $ba", 
/*0xB0*/ "BCS $*a", 
/*0xB1*/ "LDA ($a),Y", 
/*0xB2*/ "???", 
/*0xB3*/ "??? ($a),Y", 
/*0xB4*/ "LDY $a,X", 
/*0xB5*/ "LDA $a,X", 
/*0xB6*/ "LDX $a,Y", 
/*0xB7*/ "??? $a,Y", 
/*0xB8*/ "CLV", 
/*0xB9*/ "LDA $ba,Y", 
/*0xBA*/ "TSX", 
/*0xBB*/ "??? $ba,Y", 
/*0xBC*/ "LDY $ba,X", 
/*0xBD*/ "LDA $ba,X", 
/*0xBE*/ "LDX $ba,Y", 
/*0xBF*/ "??? $ba,Y", 
/*0xC0*/ "CPY #$a", 
/*0xC1*/ "CMP ($a,X)", 
/*0xC2*/ "??? #$a", 
/*0xC3*/ "??? ($a,X)", 
/*0xC4*/ "CPY $a", 
/*0xC5*/ "CMP $a", 
/*0xC6*/ "DEC $a", 
/*0xC7*/ "??? $a", 
/*0xC8*/ "INY", 
/*0xC9*/ "CMP #$a", 
/*0xCA*/ "DEX", 
/*0xCB*/ "??? #$a", 
/*0xCC*/ "CPY $ba", 
/*0xCD*/ "CMP $ba", 
/*0xCE*/ "DEC $ba", 
/*0xCF*/ "??? $ba", 
/*0xD0*/ "BNE $*a", 
/*0xD1*/ "CMP ($a),Y", 
/*0xD2*/ "???", 
/*0xD3*/ "??? ($a),Y", 
/*0xD4*/ "??? $a,X", 
/*0xD5*/ "CMP $a,X", 
/*0xD6*/ "DEC $a,X", 
/*0xD7*/ "??? $a,X", 
/*0xD8*/ "CLD", 
/*0xD9*/ "CMP $ba,Y", 
/*0xDA*/ "???", 
/*0xDB*/ "??? $ba,Y", 
/*0xDC*/ "??? $ba,X", 
/*0xDD*/ "CMP $ba,X", 
/*0xDE*/ "DEC $ba,X", 
/*0xDF*/ "??? $ba,X", 
/*0xE0*/ "CPX #$a", 
/*0xE1*/ "SBC ($a,X)", 
/*0xE2*/ "??? #$a", 
/*0xE3*/ "??? ($a,X)", 
/*0xE4*/ "CPX $a", 
/*0xE5*/ "SBC $a", 
/*0xE6*/ "INC $a", 
/*0xE7*/ "??? $a", 
/*0xE8*/ "INX", 
/*0xE9*/ "SBC #$a", 
/*0xEA*/ "NOP", 
/*0xEB*/ "??? #$a", 
/*0xEC*/ "CPX $ba", 
/*0xED*/ "SBC $ba", 
/*0xEE*/ "INC $ba", 
/*0xEF*/ "??? $ba", 
/*0xF0*/ "BEQ $*a", 
/*0xF1*/ "SBC ($a),Y", 
/*0xF2*/ "???", 
/*0xF3*/ "??? ($a),Y", 
/*0xF4*/ "??? $a,X", 
/*0xF5*/ "SBC $a,X", 
/*0xF6*/ "INC $a,X", 
/*0xF7*/ "??? $a,X", 
/*0xF8*/ "SED", 
/*0xF9*/ "SBC $ba,Y", 
/*0xFA*/ "???", 
/*0xFB*/ "??? $ba,Y", 
/*0xFC*/ "??? $ba,X", 
/*0xFD*/ "SBC $ba,X", 
/*0xFE*/ "INC $ba,X", 
/*0xFF*/ "??? $ba,X" 
}; 

}

